﻿///////////////////////////////////////////////////////////////////////////////
//
// CsvReaderクラス
//  
// Copyright (c) 2016, Advanced Core Technologies, Inc. All Rights Reserved.
//
// 2016/01/15  Rev.1.0.0.0  MOTO    作成
//
///////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;

namespace ACT.Utilities.CsvIO
{
    /// <summary>
    /// ストリームや文字列からCSVを読み取ります。
    /// </summary>
    public class CsvReader : IDisposable
    {
        //=========================================================================================
        //  Constructors
        //=========================================================================================
        #region Constructors

        /// <summary>
        /// 指定したTextReaderと見出し行の有無、区切り文字、引用符を使用して初期化します。
        /// </summary>
        /// <param name="reader">読み取りを行うTextReader</param>
        /// <param name="hasHeaderRecord">見出し行の有無。見出し行がある場合true</param>
        /// <param name="delimiter">区切り文字</param>
        /// <param name="quote">引用符</param>
        public CsvReader(TextReader reader, bool hasHeaderRecord = false,
            Delimiter delimiter = CsvIO.Delimiter.Comma,
            QuotationMark quote = CsvIO.QuotationMark.Double)
        {
            Reader = reader;
            HasHeaderRecord = hasHeaderRecord;
            Delimiter = delimiter;
            QuotationMark = quote;
            SkipEmptyLines = true;
            TrimQuotes = true;
            RemoveEscapeDoubleQuotes = true;
        }

        #endregion

        //=========================================================================================
        //  Private Constants
        //=========================================================================================
        #region Private Constants

        private static readonly char DoubleQuote = '"';

        /// <summary>
        /// レコードを分割する正規表現フォーマットです。
        /// </summary>
        private static readonly string SplitRecordPatternFormat = "(\"(?:[^\"]|\"\")*\"|[^{0}]*){0}";

        #endregion

        //=========================================================================================
        //  Private Fields
        //=========================================================================================
        #region Private Fields

        /// <summary>
        /// 見出し行の読み込み済みフラグです。
        /// </summary>
        private bool alreadyReadHeader_;

        /// <summary>
        /// Dispose済みフラグです。
        /// </summary>
        private bool alreadyDisposed_;

        #endregion

        //=========================================================================================
        //  Properties
        //=========================================================================================
        #region Properties

        /// <summary>
        /// 読み取りを行うTextReaderを取得します。
        /// </summary>
        public TextReader Reader { get; private set; }

        /// <summary>
        /// 見出し行が存在するかどうかを示す値を取得または設定します。
        /// </summary>
        public bool HasHeaderRecord { get; set; }

        /// <summary>
        /// 区切り文字を取得または設定します。
        /// </summary>
        public Delimiter Delimiter { get; set; }

        /// <summary>
        /// 区切り文字を取得します。
        /// </summary>
        public char DelimiterChar
        {
            get
            {
                char delimiter = '\0';
                switch (Delimiter)
                {
                    case Delimiter.Comma:
                        delimiter = ',';
                        break;
                    case Delimiter.Tab:
                        delimiter = '\t';
                        break;
                }

                return delimiter;
            }
        }

        /// <summary>
        /// 引用符を取得または設定します。
        /// </summary>
        public QuotationMark QuotationMark { get; set; }

        /// <summary>
        /// 空行をスキップするかどうをかを示す値を取得または設定します。
        /// スキップしない場合、空行は1つの空白フィールドを持つレコードとして読み込まれます。
        /// </summary>
        public bool SkipEmptyLines { get; set; }

        /// <summary>
        /// フィールドの先頭と末尾の引用符を削除するかどうかを示す値を取得または設定します。
        /// </summary>
        public bool TrimQuotes { get; set; }

        /// <summary>
        /// エスケープ文字を削除するかどうかを示す値を取得または設定します。
        /// </summary>
        public bool RemoveEscapeDoubleQuotes { get; set; }

        #endregion

        //=========================================================================================
        //  Public Methods
        //=========================================================================================
        #region Public Methods

        /// <summary>
        /// ストリームまたは文字列から1レコードを読み取り返します。
        /// </summary>
        /// <returns>読み取り結果のCsvRecordオブジェクト。最後まで読み取り済みの場合はnull。</returns>
        public CsvRecord ReadRecord()
        {
            string record = GetRecord(Reader, QuotationMark);
            if (record == null)
            {
                return null;
            }
            // プロパティの設定に応じて空行をスキップ
            if (SkipEmptyLines)
            {
                if (record == string.Empty)
                {
                    return ReadRecord();
                }
            }
            // 見出し行が存在し、まだ読み込んでいない場合は次のレコードを返す
            if (HasHeaderRecord)
            {
                if (!alreadyReadHeader_)
                {
                    alreadyReadHeader_ = true;
                    return ReadRecord();
                }
            }

            string[] tokens = SplitRecord(record, DelimiterChar);
            for (int i = 0; i < tokens.Length; i++)
            {
                // プロパティの設定に応じて、先頭と末尾の引用符を削除
                if (TrimQuotes)
                {
                    if (QuotationMark == CsvIO.QuotationMark.Double)
                    {
                        tokens[i] = TrimQuotationMark(tokens[i], DoubleQuote);
                    }
                }
                // プロパティの設定に応じてエスケープ文字を削除
                if (RemoveEscapeDoubleQuotes)
                {
                    tokens[i] = RemoveEscapeDoubleQuote(tokens[i]);
                }
            }

            return new CsvRecord(Delimiter, QuotationMark, tokens);
        }

        /// <summary>
        /// ストリームまたは文字列から全レコードを読み取り返します。
        /// </summary>
        /// <returns>読み取り結果のCsvRecordオブジェクト配列。</returns>
        public CsvRecord[] ReadAllRecords()
        {
            List<CsvRecord> records = new List<CsvRecord>();
            CsvRecord record;
            while ((record = ReadRecord()) != null)
            {
                records.Add(record);
            }

            return records.ToArray();
        }

        #endregion

        //=========================================================================================
        //  Private Methods
        //=========================================================================================
        #region Private Methods

        /// <summary>
        /// 引用符に囲まれた改行は無視しして、CSVの1レコードを返します。
        /// </summary>
        /// <param name="reader">読み取りを行うTextReader</param>
        /// <param name="quotationMark">引用符</param>
        /// <returns>CSVの1レコード分の文字列。</returns>
        private static string GetRecord(TextReader reader, QuotationMark quotationMark)
        {
            if (reader == null)
            {
                throw new ArgumentNullException("reader");
            }

            int quoteCount = 0;
            string concated = null;
            while (reader.Peek() != -1)
            {
                string line = reader.ReadLine();
                // 文字列を連結する
                concated = concated == null ? line : concated + Environment.NewLine + line;

                // 引用符なしの場合はループを抜ける
                if (quotationMark == QuotationMark.None)
                {
                    break;
                }
                // 引用符の数が偶数になったらループを抜ける
                quoteCount += CountChar(line, DoubleQuote);
                if ((quoteCount % 2) == 0)
                {
                    break;
                }
            }

            return concated;
        }

        /// <summary>
        /// CSVの1レコードを区切り文字で分割して返します。
        /// </summary>
        /// <param name="record">1レコード分の文字列</param>
        /// <param name="delimiter">区切り文字</param>
        /// <returns>区切り文字で分割した配列。</returns>
        private static string[] SplitRecord(string record, char delimiter)
        {
            if (record == null)
            {
                throw new ArgumentNullException(record);
            }

            // 指定区切り文字で正規表現オブジェクトを作成
            string SplitRecordPattern = string.Format(SplitRecordPatternFormat, delimiter);
            Regex splitter = new Regex(SplitRecordPattern, RegexOptions.Multiline);

            List<string> tokens = new List<string>();
            foreach (Match match in splitter.Matches(record + delimiter))
            {
                string token = match.Value;
                // 追加した末尾の区切り文字を削除して追加
                tokens.Add(token.Substring(0, token.Length - 1).Trim());
            }

            return tokens.ToArray();
        }

        /// <summary>
        /// 連続した2個のダブルクォートを1個削除します。
        /// </summary>
        /// <param name="value">対象の文字列</param>
        /// <returns>連続したダブルクォートの１つを削除した文字列。</returns>
        private static string RemoveEscapeDoubleQuote(string value)
        {
            if (value == null)
            {
                throw new ArgumentNullException("value");
            }

            return value.Replace("\"\"", "\"");
        }

        /// <summary>
        /// valueの先頭と末尾から区切り文字を削除します。
        /// </summary>
        /// <param name="value">対象の文字列</param>
        /// <param name="quote">区切り文字</param>
        /// <returns>削除した文字列。区切り文字がない場合、valueの長さが1以下の場合は元の文字列。</returns>
        private static string TrimQuotationMark(string value, char quote)
        {
            if (value == null)
            {
                throw new ArgumentNullException("value");
            }

            if (value.Length > 1 && value[0] == quote && value[value.Length - 1] == quote)
            {
                return value.Substring(1, value.Length - 2);
            }

            return value;
        }

        /// <summary>
        /// targetの文字列に含まれるvalueの数を返します。
        /// </summary>
        /// <param name="target">対象の文字列</param>
        /// <param name="value">検索する文字</param>
        /// <returns>valueが含まれている数。targetまたはvalueがnullの時は0。</returns>
        private static int CountChar(string target, char value)
        {
            if (target == null)
            {
                return 0;
            }

            return target.Length - target.Replace(value.ToString(), string.Empty).Length;
        }

        #endregion

        //=========================================================================================
        //  IDisposable
        //=========================================================================================
        #region IDisposable

        /// <summary>
        /// このCsvReaderオブジェクトによって使用されている全てのリソースを解放します。
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// 全てのリソースを解放します。
        /// </summary>
        /// <param name="disposing"></param>
        protected virtual void Dispose(bool disposing)
        {
            if (alreadyDisposed_)
            {
                return;
            }

            if (disposing)
            {
                // マネージリソースの解放
                Reader.Dispose();

            }
            // アンマネージリソースの解放

            alreadyDisposed_ = true;
        }

        /// <summary>
        /// デストラクターです。
        /// </summary>
        ~CsvReader()
        {
            Dispose(false);
        }

        #endregion
    }
}
