﻿///////////////////////////////////////////////////////////////////////////////////////////////////
// 
//  ACT距離計算サービス サンプルプログラム
//  MainFormクラス
//
//  Copyrights (c) 2016-2023, Advanced Core Technologies, Inc. All rights reserved.
//
//  2023/09/21 Rev.1.0.0.1 KYON AdminURLに合わせてSSL使用フラグを変更出来るように変更
//  2016/01/28 Rev.1.0.0.0 MOTO 作成
//
///////////////////////////////////////////////////////////////////////////////////////////////////

using ACT.Utilities.CsvIO;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace ACTSOAPCSSearchAddress
{
    public partial class MainForm : Form
    {
        //=========================================================================================
        //  Constructors
        //=========================================================================================
        #region Constructors

        /// <summary>
        /// 既定値を使用して初期化します。
        /// </summary>
        public MainForm()
        {
            InitializeComponent();
        }

        #endregion

        //=========================================================================================
        //  Private Fields
        //=========================================================================================
        #region Private Fields

        /// <summary>
        /// ユーザー情報です。
        /// </summary>
        private UserSettings userSettings_;

        /// <summary>
        /// 距離計算サービスのユーザーです。
        /// </summary>
        private User user_;

        #endregion

        //=========================================================================================
        //  Private Constants
        //=========================================================================================
        #region Private Constants

        /// <summary>
        /// 地点名の列番号です。
        /// </summary>
        private static readonly int NameColumnIndex = 0;

        /// <summary>
        /// 住所の列番号です。
        /// </summary>
        private static readonly int AddressColumnIndex = 1;

        #endregion

        //=========================================================================================
        //  Private Methods
        //=========================================================================================
        #region Private Methods

        /// <summary>
        /// 検索を実行し実行結果をファイルに出力します。
        /// </summary>
        /// <param name="userSettings">ユーザー設定</param>
        /// <returns>実行結果のメッセージ。</returns>
        private string DoSearch(UserSettings userSettings)
        {
            StringBuilder resultMessage = new StringBuilder();

            try
            {
                //---------------------------------------------------------------------------------
                //  管理サービスオブジェクトと環境設定オブジェクトの作成
                //---------------------------------------------------------------------------------
                // 管理サービスオブジェクトを作成
                DCAdmin.ACTDCADMIN adminService = new DCAdmin.ACTDCADMIN();
                // エンドポイントURLを設定
                adminService.Url = user_.AdminUrl;

                // 環境設定オブジェクトを作成
                DCAdmin.Environ adminEnviron = user_.GetAdminEnviron();

                //---------------------------------------------------------------------------------
                //  ユーザー情報オブジェクトの取得
                //---------------------------------------------------------------------------------
                // ユーザー情報オブジェクトを作成
                DCAdmin.UserInformation2 userInfo = new DCAdmin.UserInformation2();
                // ユーザー情報取得サービスの呼び出し
                if (!adminService.GetUserInformation2(adminEnviron, ref userInfo))
                {
                    // ユーザー情報の取得に失敗した場合
                    resultMessage.AppendLine("ユーザー情報の取得に失敗しました。");
                    return resultMessage.ToString();
                }
                user_.UserInfo = userInfo;

                //---------------------------------------------------------------------------------
                //  住所検索サービスオブジェクトと環境設定オブジェクトの作成
                //---------------------------------------------------------------------------------
                GCService.ACTGCWS gcService = new GCService.ACTGCWS();
                // 住所検索サーバー取得サービスの呼び出し
                if (!adminService.GetUsableGeoServer(ref adminEnviron))
                {
                    // 距離計算サーバーの取得に失敗した場合
                    resultMessage.AppendLine("住所検索サーバーの情報取得に失敗しました。");
                    return resultMessage.ToString();
                }
                // エンドポイントURLを設定
                gcService.Url = Utilities.ChangeProtocol(adminEnviron.WSURL, CalcSettings.UseSsl);
                resultMessage.AppendLine("住所検索サーバー: " + gcService.Url +
                    "(" + adminEnviron.Host + ":" + adminEnviron.Port + ")");

                // 環境設定構造体オブジェクトを作成
                GCService.Environ gcEnviron = Utilities.GetGCServiceEnviron(adminEnviron);

                //---------------------------------------------------------------------------------
                //  CSVファイルを読み込み住所検索アイテム構造体配列を作成
                //---------------------------------------------------------------------------------
                // ファイルから住所を読み込む
                CsvRecord[] records = ReadCsvFile(userSettings.InputFilePath);
                if (records == null)
                {
                    // CSVファイルの読み込みに失敗した場合
                    resultMessage.AppendLine("入力ファイルの読み込みに失敗しました。");
                    return resultMessage.ToString();
                }

                // 住所検索アイテム構造体に読み取った経度緯度情報を設定
                GCService.SearchAddressItem[] items = CreateItems(records);

                //---------------------------------------------------------------------------------
                //  検索住所レベルの設定
                //---------------------------------------------------------------------------------
                // ユーザーの最大検索住所レベルを検索住所レベルに設定
                int searchAddressLevel = user_.UserInfo.GeoMaxAddressLevel;
                resultMessage.AppendLine("検索住所レベル: " + searchAddressLevel);

                //---------------------------------------------------------------------------------
                //  共通パラメータ構造体の設定
                //---------------------------------------------------------------------------------
                // 共通パラメータ構造体オブジェクトを作成
                GCService.CommonParam param = new GCService.CommonParam()
                {
                    Datum = (int)CalcSettings.Datum,
                    CoordinateFormat = CalcSettings.CoordinateFormat,
                    SearchFlag = (int)CalcSettings.SearchFlag,
                };

                //---------------------------------------------------------------------------------
                //  住所検索の実行＆実行結果の取得
                //---------------------------------------------------------------------------------
                // 検索結果を格納する変数を作成
                GCService.GeoInfo geoInfo = null;

                // 検索可能地点数を取得
                int maxLocs = user_.UserInfo.GeoMaxLocs;

                // 必要検索回数を求める
                int numSearch = (int)Math.Ceiling((double)items.Length / (double)maxLocs);

                // 分割して検索を実行
                for (int i = 0; i < numSearch; i++)
                {
                    // 対象地点を設定
                    int start = maxLocs * i;
                    int length = i + 1 == numSearch ? items.Length - start : maxLocs;
                    GCService.SearchAddressItem[] target =
                        items.Copy<GCService.SearchAddressItem>(start, length);
                    
                    // 住所検索の実行
                    int result = gcService.SearchAddress(
                        gcEnviron, param, searchAddressLevel, ref target, ref geoInfo);

                    // 実行結果の取得
                    if (result == 0)
                    {
                        // 成功
                        resultMessage.AppendLine(
                            string.Format("{0} / {1} 回目：成功", i + 1, numSearch));
                    }
                    else
                    {
                        // 失敗、エラーメッセージを取得
                        resultMessage.AppendLine(
                            string.Format("{0} / {1} 回目：失敗 {2}{3}", i + 1, numSearch,
                            Environment.NewLine, gcService.GetGCWSMessage(gcEnviron, result)));
                    }

                    // 実行結果を元の配列に書き戻す
                    Array.Copy(target, 0, items, start, length);

                    // 最後のループ以外は待機
                    if (i + 1 != numSearch)
                    {
                        // ミリ秒単位の待機時間を取得
                        int waitTime = user_.UserInfo.GeoInterval * 1000;
                        // 待機
                        System.Threading.Thread.Sleep(waitTime);
                    }
                }

                // 最終住所検索日時を更新
                user_.UpdateLastGCTime();

                //---------------------------------------------------------------------------------
                //  ファイル出力
                //---------------------------------------------------------------------------------
                if (!WriteItems(userSettings.OutputFilePath, items))
                {
                    // 出力に失敗した場合
                    resultMessage.AppendLine("ファイルの書き込みに失敗しました。");
                }
            }
            catch (System.Net.WebException e)
            {
                resultMessage.AppendLine("通信エラーが発生しました。");
                resultMessage.AppendLine(e.Message);
            }
            catch
            {
                //
            }

            return resultMessage.ToString();
        }

        /// <summary>
        /// 検索を実行し実行結果をファイルに出力します。
        /// </summary>
        /// <param name="userSettings"></param>
        /// <returns>実行結果のメッセージ。</returns>
        private string DoWork(UserSettings userSettings)
        {
            string resultMessage = string.Empty;
            // 進捗フォームを表示して検索を実行
            ProgressDialog.Run(() => resultMessage = DoSearch(userSettings), "検索中");

            return resultMessage;
        }

        /// <summary>
        /// CSVファイルの全レコードを読み取り結果を返します。
        /// </summary>
        /// <param name="path">対象ファイルのパス</param>
        /// <returns>読み取り結果のCsvRecord配列。読み取りに失敗した場合はnull。</returns>
        private static CsvRecord[] ReadCsvFile(string path)
        {
            CsvRecord[] records = null;
            try
            {
                using (StreamReader strmReader =
                    new StreamReader(path, Encoding.GetEncoding("shift_jis")))
                using (CsvReader csvReader =
                    new CsvReader(strmReader, false, Delimiter.Comma, QuotationMark.Double))
                {
                    records = csvReader.ReadAllRecords();
                }
            }
            catch
            {
                //
            }

            return records;
        }

        /// <summary>
        /// CsvRecord配列から住所検索アイテム構造体配列を作成し返します。
        /// </summary>
        /// <param name="records">CsvRecord配列</param>
        /// <returns>住所検索アイテム構造体配列。</returns>
        private static GCService.SearchAddressItem[] CreateItems(CsvRecord[] records)
        {
            // 住所検索アイテム構造体に住所と地点名を設定
            GCService.SearchAddressItem[] items = new GCService.SearchAddressItem[records.Length];
            for (int i = 0; i < records.Length; i++)
            {
                items[i] = new GCService.SearchAddressItem();
                items[i].SearchAddress = records[i][AddressColumnIndex];
                items[i].Tag = records[i][NameColumnIndex];
            }

            return items;
        }

        /// <summary>
        /// 住所検索アイテム構造体をCSVファイルに書き込みます。
        /// </summary>
        /// <param name="path">出力ファイルのパス</param>
        /// <param name="items">計算結果を格納した住所検索アイテム構造体配列</param>
        /// <returns>書き込みに成功した場合はtrue、失敗した場合はfalse。</returns>
        private static bool WriteItems(string path, IEnumerable<GCService.SearchAddressItem> items)
        {
            try
            {
                using (StreamWriter strmWriter =
                    new StreamWriter(path, false, Encoding.GetEncoding("shift_jis")))
                using (CsvWriter csvWriter = new CsvWriter(strmWriter))
                {
                    foreach (var item in items)
                    {
                        CsvRecord record = new CsvRecord(Delimiter.Comma, QuotationMark.Double);
                        record.Add(item.Tag);
                        record.Add(item.SearchAddress);
                        record.Add(item.Lon.ToString());
                        record.Add(item.Lat.ToString());

                        csvWriter.WriteRecord(record);
                    }
                }
            }
            catch
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// コントロールからユーザー設定を取得します。
        /// </summary>
        /// <returns>取得したUserSettingsオブジェクト。</returns>
        private UserSettings GetUserSettingsFromControls()
        {
            UserSettings settings = new UserSettings()
            {
                UserID = textBoxUserId.Text,
                Password = textBoxPassword.Text,
                AdminUrl = textBoxAdminServerUrl.Text,
                InputFilePath = textBoxInputFilePath.Text,
                OutputFilePath = textBoxOutputFilePath.Text,
            };

            return settings;
        }

        /// <summary>
        /// 指定したユーザー設定をコントロールに適用します。
        /// </summary>
        /// <param name="settings">適用するUserSettingsオブジェクト</param>
        private void ApplyUserSettingsToControls(UserSettings settings)
        {
            textBoxUserId.Text = settings.UserID;
            textBoxAdminServerUrl.Text = settings.AdminUrl;
            textBoxInputFilePath.Text = settings.InputFilePath;
            textBoxOutputFilePath.Text = settings.OutputFilePath;
        }

        /// <summary>
        /// 入力ファイルを選択し、選択したファイルのパスを返します。
        /// </summary>
        /// <param name="oldPath">現在の入力ファイルパス</param>
        /// <returns>選択したファイルのパス。キャンセルの場合はnull。</returns>
        private static string SelectInputFile(string oldPath)
        {
            string newPath = null;
            using (OpenFileDialog dialog = new OpenFileDialog())
            {
                if (File.Exists(oldPath))
                {
                    dialog.InitialDirectory = Path.GetDirectoryName(oldPath);
                    dialog.FileName = Path.GetFileName(oldPath);
                }
                dialog.Title = "入力ファイルの選択";
                dialog.Filter = "CSVファイル (*.csv)|*.csv|全てのファイル (*.*)|*.*";

                if (dialog.ShowDialog() == DialogResult.OK)
                {
                    newPath = dialog.FileName;
                }
            }

            return newPath;
        }

        /// <summary>
        /// 出力ファイルを選択し、選択したファイルのパスを返します。
        /// </summary>
        /// <param name="oldPath">現在の出力ファイルパス</param>
        /// <returns>選択したファイルのパス。キャンセルの場合はnull。</returns>
        private static string SelectOutputFile(string oldPath)
        {
            string newPath = null;
            using (SaveFileDialog dialog = new SaveFileDialog())
            {
                if (File.Exists(oldPath))
                {
                    dialog.InitialDirectory = Path.GetDirectoryName(oldPath);
                    dialog.FileName = Path.GetFileName(oldPath);
                }
                dialog.Title = "出力ファイルの選択";
                dialog.Filter = "CSVファイル (*.csv)|*.csv|全てのファイル (*.*)|*.*";

                if (dialog.ShowDialog() == DialogResult.OK)
                {
                    newPath = dialog.FileName;
                }
            }

            return newPath;
        }

        /// <summary>
        /// ユーザー設定を検証します。
        /// </summary>
        /// <param name="userID">ユーザーID</param>
        /// <param name="password">パスワード</param>
        /// <param name="adminUrl">管理サーバーURL</param>
        /// <param name="inputFilePath">入力ファイルのパス</param>
        /// <param name="outputFilePath">出力ファイルのパス</param>
        /// <returns>設定が全て有効な場合はnull、それ以外の場合はエラーメッセージ。</returns>
        private static string ValidateSettings(string userID, string password, string adminUrl,
            string inputFilePath, string outputFilePath)
        {
            StringBuilder message = new StringBuilder();
            if (string.IsNullOrEmpty(userID))
            {
                message.AppendLine("ユーザーIDが指定されていません。");
            }
            if (string.IsNullOrEmpty(password))
            {
                message.AppendLine("パスワードが指定されていません。");
            }
            if (string.IsNullOrEmpty(adminUrl))
            {
                message.AppendLine("管理サーバーURLが指定されていません。");
            }
            string inputFilePathError = ValidateInputFilePath(inputFilePath);
            if (inputFilePathError != null)
            {
                message.AppendLine(inputFilePathError);
            }
            string outputFilePathError = ValidateOutputFilePath(outputFilePath);
            if (outputFilePathError != null)
            {
                message.AppendLine(outputFilePathError);
            }
            if (message.Length == 0)
            {
                if (ConfirmOverwrite(outputFilePath))
                {
                    return null;
                }
                message.AppendLine("指定された " + outputFilePath + " は既に存在します。");
            }

            return message.ToString();
        }

        /// <summary>
        /// 入力ファイルのパスを検証します。
        /// </summary>
        /// <param name="path">検査対象のパス</param>
        /// <returns>有効な場合はnull、それ以外の場合はエラーメッセージ。</returns>
        private static string ValidateInputFilePath(string path)
        {
            if (string.IsNullOrEmpty(path))
            {
                return "入力ファイルが指定されていません。";
            }
            try
            {
                if (!File.Exists(path))
                {
                    return "指定された " + path + " が存在しません。";
                }
            }
            catch
            {
                return "入力ファイルのパスが不正です。";
            }

            return null;
        }

        /// <summary>
        /// 出力ファイルのパスを検証します。
        /// </summary>
        /// <param name="path">検査対象のパス</param>
        /// <returns>有効な場合はnull、それ以外の場合はエラーメッセージ。</returns>
        private static string ValidateOutputFilePath(string path)
        {
            if (string.IsNullOrEmpty(path))
            {
                return "出力ファイルが指定されていません。";
            }
            try
            {
                string directoryName = Path.GetDirectoryName(path);
                if (string.IsNullOrEmpty(directoryName))
                {
                    return "指定されたフォルダが存在しません。";
                }
                if (!Directory.Exists(directoryName))
                {
                    return "指定された " + directoryName + " が存在しません。";
                }
            }
            catch
            {
                return "出力ファイルのパスが不正です。";
            }

            return null;
        }

        /// <summary>
        /// ファイルが既に存在する場合に上書き確認を行います。
        /// </summary>
        /// <param name="path">検査対象のパス</param>
        /// <returns>ファイルが既に存在していて上書きしない場合はtrue、それ以外の場合はfalse。</returns>
        private static bool ConfirmOverwrite(string path)
        {
            if (File.Exists(path))
            {
                DialogResult result = MessageBox.Show(path + " は既に存在しています。" +
                    Environment.NewLine + "上書きしてよろしいですか？", "上書き確認",
                    MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation,
                    MessageBoxDefaultButton.Button2);
                if (result == DialogResult.No)
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// 再検索可能までの残待機時間を更新します。
        /// </summary>
        /// <param name="waitingSeconds">秒単位の残り待機時間</param>
        private void UpdateWaitingTime(int waitingSeconds)
        {
            string labelText = string.Empty;
            if (waitingSeconds > 0)
            {
                // 待機時間が経過していない場合
                labelText = string.Format("再検索可能まで {0:00} 秒", waitingSeconds);
            }
            // コントロールの更新
            labelCountDown.Text = labelText;
            buttonSearch.Enabled = waitingSeconds <= 0;
            timerCalcWait.Enabled = waitingSeconds > 0;
        }

        #endregion

        //=========================================================================================
        //  Event Handlers
        //=========================================================================================
        #region Event Handlers

        /// <summary>
        /// メインフォームのロードイベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            // 保存したユーザー設定を復元し設定する
            userSettings_ = UserSettings.Load(UserSettings.SettingFilePath);
            if (userSettings_ == null)
            {
                // 失敗した場合は既定値を設定
                userSettings_ = new UserSettings();
                textBoxAdminServerUrl.Text = CalcSettings.DefaultAdminUrl;
                return;
            }
            ApplyUserSettingsToControls(userSettings_);
        }

        /// <summary>
        /// メインフォームのクローズ後イベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            // ユーザー設定を保存する
            userSettings_ = GetUserSettingsFromControls();
            try
            {
                userSettings_.Save(UserSettings.SettingFilePath);
            }
            catch
            {
                //
            }
        }

        /// <summary>
        /// 「既定値」ボタンのクリックイベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonResetUrl_Click(object sender, EventArgs e)
        {
            textBoxAdminServerUrl.Text = CalcSettings.DefaultAdminUrl;
        }

        /// <summary>
        /// 「参照」ボタンのクリックイベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonSelectInputFile_Click(object sender, EventArgs e)
        {
            string path = SelectInputFile(textBoxInputFilePath.Text) ?? textBoxInputFilePath.Text;
            textBoxInputFilePath.Text = path;
        }

        /// <summary>
        /// 「参照」ボタンのクリックイベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonSelectOutputFile_Click(object sender, EventArgs e)
        {
            string path = SelectOutputFile(textBoxOutputFilePath.Text) ?? textBoxOutputFilePath.Text;
            textBoxOutputFilePath.Text = path;
        }
        
        /// <summary>
        /// 「検索」ボタンのクリックイベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonSearch_Click(object sender, EventArgs e)
        {
            // フォームから設定を取得し検証
            userSettings_ = GetUserSettingsFromControls();
            string errorMessage = ValidateSettings(userSettings_.UserID, userSettings_.Password,
                userSettings_.AdminUrl, userSettings_.InputFilePath, userSettings_.OutputFilePath);
            if (errorMessage != null)
            {
                textBoxResults.Text = errorMessage;
                return;
            }

            // テキストボックスの表示を変更
            textBoxResults.Text = "検索中...";

            // 「検索」ボタンを無効
            buttonSearch.Enabled = false;

            // AdminURLを確認してSSL使用フラグを変更
            CalcSettings.UseSsl = userSettings_.AdminUrl.StartsWith("https") == true;

            // ユーザーを作成
            user_ = new User(userSettings_.UserID, userSettings_.Password,
                userSettings_.AdminUrl, CalcSettings.UseSsl);

            // 検索の実行
            textBoxResults.Text = DoWork(userSettings_);

            // 再検索までのカウントダウン
            timerCalcWait.Start();
        }

        /// <summary>
        /// 「終了」ボタンのクリックイベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        /// <summary>
        /// 「マニュアルを開く」リンクのクリックイベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void linkLabelManual_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            try
            {
                Process.Start(UserSettings.UsersGuideFilePath);
            }
            catch
            {
                //
            }
        }

        /// <summary>
        /// 計算待機時間タイマーのイベントです。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void timerCalcWait_Tick(object sender, EventArgs e)
        {
            if (user_.UserInfo == null)
            {
                UpdateWaitingTime(-1);
                return;
            }
            int elpsedSeconds = (int)((DateTime.Now - user_.LastGCTime).TotalSeconds);
            int waitingSeconds = user_.UserInfo.GeoInterval - elpsedSeconds;
            UpdateWaitingTime(waitingSeconds);
        }

        #endregion
    }
}
